/* This was primarily taken from a file written by Mark Allen Weiss
 * (available at http://www.cs.fiu.edu/~weiss/dsaajava/Code/DataStructures/BinaryHeap.java)
 * It was then modified by Ed Hong to conform to the priority queue ADT in
 * Goodrich's Algorthims book. This implements a minimizing binary-Heap based
 * Priority Queue, where the key is always a double. This does not
 * implement the extension of the PQ ADT (so that remove(), replaceKey(),
 * and locators are not implemented. Also note that this priority does not
 * use comparators. This implementation may not handle errors well.
 * Modified 2/14/05.
 * The relevant implemented calls are
 	public void insertItem(double k, Object o);  -> associate key k with object o
	public Object removeMin();                   -> remove and return object with minimum key
    public double minKey();						 -> return minimum key present in pq
	public Object minElement();                  -> return object associated with min key (do not remove)
	public int size();							 -> return # items in pq
	public boolean isEmpty();                    -> check if heap is empty.

	public BinaryHeapArrayPQ( int capacity );  -> create new pq with given capacity
 */
//
// BinaryHeap class
// CONSTRUCTION: with optional capacity (that defaults to 100)
//
// ******************PUBLIC OPERATIONS*********************
// (for using binary heap directly)
// void insert( x )       --> Insert x
// Comparable deleteMin( )--> Return and remove smallest item
// Comparable findMin( )  --> Return smallest item
// boolean isEmpty( )     --> Return true if empty; else false
// boolean isFull( )      --> Return true if full; else false
// void makeEmpty( )      --> Remove all items
// for using the PQ ADT, there are now insertItem and removeMin operations
//
// ******************ERRORS********************************
// Throws Overflow if capacity exceeded

/**
 * Implements a binary heap.
 * Note that all "matching" is based on the compareTo method.
 * @author Mark Allen Weiss
 */
public class BinaryHeapArrayPQ
    {
        /**
         * Construct the binary heap.
         */
        public BinaryHeapArrayPQ( )
        {
            this( DEFAULT_CAPACITY );
        }

        /**
         * Construct the binary heap.
         * @param capacity the capacity of the binary heap.
         */
        public BinaryHeapArrayPQ( int capacity )
        {
            currentSize = 0;
            array = new Comparable[ capacity + 1 ];
        }

	// Store (key,object) pairs in priority queue;
	// make them comparable so that the BinaryHeap implementation
	// from below can be used.
		private class Keyobjectpair implements Comparable{
			public double key;
			public Object element;

			public Keyobjectpair(double k, Object o) {
				key = k;
				element = o;
			}
			public int compareTo(Object o) {
				Keyobjectpair o2 = (Keyobjectpair) o;
				if (this.key < o2.key ) return -1;
				else if (this.key == o2.key) return 0;
				else return 1;
			}
		}

/**
 * associate key k with object o
 */
		public void insertItem(double k, Object o) throws Overflow{
			insert(new Keyobjectpair(k,o));
		}

/**
 * remove and return object with minimum key
 */
		public Object removeMin() {
			return ((Keyobjectpair) deleteMin()).element;
		}

/**
 * return minimum key present in pq
 */
		public double minKey() {
			return ((Keyobjectpair) findMin()).key;
		}

/**
 *return object associated with min key (do not remove)
 */
		public Object minElement() {
			return ((Keyobjectpair) findMin()).element;
		}

/**
 * return # items in pq
 */
		public int size() {
		  return currentSize;
		}

        /**
         * Insert into the priority queue, maintaining heap order.
         * Duplicates are allowed.
         * @param x the item to insert.
         * @exception Overflow if container is full.
         */
        public void insert( Comparable x ) throws Overflow
        {
            if( isFull( ) )
                throw new Overflow( );

	    // Percolate up
            int hole = ++currentSize;
            for( ; hole > 1 && x.compareTo( array[ hole / 2 ] ) < 0; hole /= 2 )
                array[ hole ] = array[ hole / 2 ];
            array[ hole ] = x;
        }

        /**
         * Find the smallest item in the priority queue.
         * @return the smallest item, or null, if empty.
         */
        public Comparable findMin( )
        {
            if( isEmpty( ) )
                return null;
            return array[ 1 ];
        }

        /**
         * Remove the smallest item from the priority queue.
         * @return the smallest item, or null, if empty.
         */
        public Comparable deleteMin( )
        {
            if( isEmpty( ) )
                return null;

            Comparable minItem = findMin( );
            array[ 1 ] = array[ currentSize-- ];
            percolateDown( 1 );

            return minItem;
        }

        /**
         * Establish heap order property from an arbitrary
         * arrangement of items. Runs in linear time.
         */
        private void buildHeap( )
        {
            for( int i = currentSize / 2; i > 0; i-- )
                percolateDown( i );
        }

        /**
         * Test if the priority queue is logically empty.
         * @return true if empty, false otherwise.
         */
        public boolean isEmpty( )
        {
            return currentSize == 0;
        }

        /**
         * Test if the priority queue is logically full.
         * @return true if full, false otherwise.
         */
        public boolean isFull( )
        {
            return currentSize == array.length - 1;
        }

        /**
         * Make the priority queue logically empty.
         */
        public void makeEmpty( )
        {
            currentSize = 0;
        }

        private static final int DEFAULT_CAPACITY = 100;

        private int currentSize;      // Number of elements in heap
        private Comparable [ ] array; // The heap array

        /**
         * Internal method to percolate down in the heap.
         * @param hole the index at which the percolate begins.
         */
        private void percolateDown( int hole )
        {
	    /* 1*/      int child;
	    /* 2*/      Comparable tmp = array[ hole ];

	    /* 3*/      for( ; hole * 2 <= currentSize; hole = child )
		{
		    /* 4*/          child = hole * 2;
		    /* 5*/          if( child != currentSize &&
					/* 6*/                  array[ child + 1 ].compareTo( array[ child ] ) < 0 )
			/* 7*/              child++;
		    /* 8*/          if( array[ child ].compareTo( tmp ) < 0 )
			/* 9*/              array[ hole ] = array[ child ];
		    else
			/*10*/              break;
		}
	    /*11*/      array[ hole ] = tmp;
        }

	// Test program
        public static void main( String [ ] args )
        {
            int numItems = 10000;
            BinaryHeapArrayPQ h = new BinaryHeapArrayPQ( numItems );
            int i = 37;

            try
		{
		    for( i = 37; i != 0; i = ( i + 37 ) % numItems )
			h.insert( new Integer( i ) );
		    for( i = 1; i < numItems; i++ )
			if( ((Integer)( h.deleteMin( ) )).intValue( ) != i )
			    System.out.println( "Oops! " + i );

		    for( i = 37; i != 0; i = ( i + 37 ) % numItems )
			h.insert( new Integer( i ) );
		    h.insert( new Integer( 0 ) );
		    i = 9999999;
		    h.insert( new Integer( i ) );
		    for( i = 1; i <= numItems; i++ )
			if( ((Integer)( h.deleteMin( ) )).intValue( ) != i )
			    System.out.println( "Oops! " + i + " " );
		}
            catch( Overflow e )
		{ System.out.println( "Overflow (expected)! " + i  ); }
        }
    }
