/*
Written by Ed Hong
UWT, Feb 28, 2005.
 * This implements a simple Partition ADT with the Union-Find operations, using
 * locators, and LinkedLists to store the sets. It is based on the
 * ADT is described in Goodrich, section 4.2.2. Note that sets here are basically
 * linkedlists of locators.
 * The relevant implemented calls are
 	public Locator makeSet(Object o);
 	       -> Create a singleton set containing one
 	           object o; return a locator to the object.
	public void Union (LinkedList A, LinkedList B);
	       -> Adds all items from set B into set A.
	          updates B's items to point to "this".
	          sets B to be an empty set.
	          (In other words: this <- this Union B; B <- null)
	public LinkedList find(Locator l)
	       -> returns the set containing the item that Locator l points to.

    Useful locator calls:
      public Locator.element() -> returns the object that the locator points to.
*/
import java.util.*;


public class SimplePartition {

   // internal class for containing (containing set, item, locator)
   // no methods beside constructor.
  private class setitem {
      public Object element;  // element inserted
      public LinkedList parent; // reference to set that this item is in.
      public PartitionLocator loc; // reference to locator
      // create a new item with parent s, object o.
      public setitem(LinkedList s, Object o) {
		parent = s;
		element = o;
		loc = new PartitionLocator(this); // initialize locator to point to this item
      }
  }

  private class PartitionLocator implements Locator {
      private setitem locatordat; // reference to item inside Partition
      // constructor..
      private PartitionLocator(setitem k) {
		locatordat = k;
      }
      public double key() {
		// ParitionLocators have no associated key.
		// print error message and return.
		System.out.println("Error: Should not call PartitionLocator.key()");
 		return -1;
      }
      public Object element() {
	  /* returns Object associated to referenced data */
		return locatordat.element;
      }
      public String toString() {
		  return locatordat.element.toString();
	  }
  }

    private LinkedList listOfSets; // contains a list of each set ever created
                                   // for debugging.

    /** create a new partition */
	public SimplePartition() {
		listOfSets = new LinkedList();
    }

/**
 	public Locator makeSet(Object o);
 	       -> Create a singleton set containing one
 	           object o; return a locator to the object.
 *
 */
	public Locator makeSet(Object o) {
	    LinkedList newList = new LinkedList();
	    setitem newItem = new setitem(newList, o);
	    newList.add(newItem.loc);
	    listOfSets.add(newList);
	    return newItem.loc;
	}

/**
 *
	public void Union (LinkedList A, LinkedList B);
	       -> Adds all items from set B into set A.
	          updates  B's items to point to "this".
	          sets B to be an empty set.
	          (In other words: this <- this Union B; B <- null)
	 gives errors if sets A and B are not created by SimplePartition.
	 You should pass in LinkedLists from find(loc) calls, where loc is a locator
	 returned by a makeSet call.
 */
    public void Union (LinkedList A, LinkedList B) {
		if (A == B) return;
		LinkedList C;
		// make B hold the smaller set.
		if (A.size () < B.size ()) {
			C = A;
			A = B;
			B = C;
		}
		// now insert items from B to A.
        Iterator it = B.iterator();
        while (it.hasNext()) {
			PartitionLocator temp = (PartitionLocator) it.next();
			temp.locatordat.parent = A;  // make this item point to A.
			A.add(temp); // put it on list A
		}
		// empty out list B
		B.clear();
	}
/**
	public LinkedList find(Locator l)
	       -> returns the set containing the item that Locator l points to.
 */
	public LinkedList find(Locator l) {
         return ((PartitionLocator) l).locatordat.parent;
	}

	// Test program
        public static void main( String [ ] args )
        {
            int numItems = 11;
            SimplePartition h = new SimplePartition( );
	    java.util.Random r = new java.util.Random();
	    int i, temp;
	    Locator[] k = new Locator[numItems];
		for( i = 0; i < numItems; i++ ) {
			temp = r.nextInt(1000);
			k[i]= h.makeSet( "num" + temp  );
			System.out.println("makeSet num" + temp);
		}
		for( i = 0; i < numItems; i++ ) {
			temp = (i*r.nextInt(1000))%numItems;
		    System.out.println("Union of " + k[i] + " with " + k[temp]);
			h.Union(h.find(k[i]), h.find(k[temp]));
			System.out.println("Items now in following sets: ");
			for (int j = 0; j < numItems; j++)
			   System.out.println( j + " " + h.find(k[j]));
		}
	}

}

